// Board class
// CSC 375 - Lea
// Jacob Peck


import java.util.*;

public class Board {
  /* cells (2 bits per cell)                cell types
     | 00 |    | 01 |    | 02 |    | 03          a
  --------------------------------------
  04 |    | 05 |    | 06 |    | 07 |             b
  --------------------------------------
     | 08 |    | 09 |    | 10 |    | 11          a
  --------------------------------------
  12 |    | 13 |    | 14 |    | 15 |             b
  --------------------------------------
     | 16 |    | 17 |    | 18 |    | 19          a
  --------------------------------------
  20 |    | 21 |    | 22 |    | 23 |             b
  --------------------------------------
     | 24 |    | 25 |    | 26 |    | 27          a
  --------------------------------------
  28 |    | 29 |    | 30 |    | 31 |             b
  */
  
  /* cell contents:
     00 = empty
     01 = red
     10 = black
  */
  
  // king mask = 32 bit integer, each bit 0 = not king, 1 = king
  
  public static final boolean KING = true;
  public static final boolean PAWN = false;
  public static final int EMPTY = 0;
  public static final int RED = 1;
  public static final int BLACK = 2;
  public static final int ERROR = 3; // error state, 0b11 in a cell in contents
  private static boolean initialized = false;
  private static long instancecount = 0;
  
  public static final HashMap<Integer,Integer[]> CAN_MOVE_LIST = new HashMap<Integer,Integer[]>();
  
  private static void initializeMovePermissions() {
    CAN_MOVE_LIST.put(0,new Integer[]{4,5,9});
    CAN_MOVE_LIST.put(1,new Integer[]{5,6,8,10});
    CAN_MOVE_LIST.put(2,new Integer[]{6,7,9,11});
    CAN_MOVE_LIST.put(3,new Integer[]{7,10});
    
    CAN_MOVE_LIST.put(4,new Integer[]{0,8,13});
    CAN_MOVE_LIST.put(5,new Integer[]{0,1,8,9,12,14});
    CAN_MOVE_LIST.put(6,new Integer[]{1,2,9,10,13,15});
    CAN_MOVE_LIST.put(7,new Integer[]{2,3,10,11,14});
    
    CAN_MOVE_LIST.put(8,new Integer[]{1,4,5,12,13,17});
    CAN_MOVE_LIST.put(9,new Integer[]{0,2,5,6,13,14,16,18});
    CAN_MOVE_LIST.put(10,new Integer[]{1,3,6,7,14,15,17,19});
    CAN_MOVE_LIST.put(11,new Integer[]{2,7,15,18});
    
    CAN_MOVE_LIST.put(12,new Integer[]{5,8,16,21});
    CAN_MOVE_LIST.put(13,new Integer[]{4,6,8,9,16,17,20,22});
    CAN_MOVE_LIST.put(14,new Integer[]{5,7,9,10,17,18,21,23});
    CAN_MOVE_LIST.put(15,new Integer[]{6,10,11,18,19,22});
    
    CAN_MOVE_LIST.put(16,new Integer[]{9,12,13,20,21,25});
    CAN_MOVE_LIST.put(17,new Integer[]{8,10,13,14,21,22,24,26});
    CAN_MOVE_LIST.put(18,new Integer[]{9,11,14,15,22,23,25,27});
    CAN_MOVE_LIST.put(19,new Integer[]{10,15,23,26});
    
    CAN_MOVE_LIST.put(20,new Integer[]{13,16,24,29});
    CAN_MOVE_LIST.put(21,new Integer[]{12,14,16,17,24,25,28,30});
    CAN_MOVE_LIST.put(22,new Integer[]{13,15,17,18,25,26,29,31});
    CAN_MOVE_LIST.put(23,new Integer[]{14,18,19,26,27,30});
    
    CAN_MOVE_LIST.put(24,new Integer[]{17,20,21,28,29});
    CAN_MOVE_LIST.put(25,new Integer[]{16,18,21,22,29,30});
    CAN_MOVE_LIST.put(26,new Integer[]{17,19,22,23,30,31});
    CAN_MOVE_LIST.put(27,new Integer[]{18,23,31});
    
    CAN_MOVE_LIST.put(28,new Integer[]{21,24});
    CAN_MOVE_LIST.put(29,new Integer[]{20,22,24,25});
    CAN_MOVE_LIST.put(30,new Integer[]{21,23,25,26});
    CAN_MOVE_LIST.put(31,new Integer[]{22,26,27});
  }
  
  
  private long contents = 0; // empty board
  private int kingmask = 0; // no kings
  
  public Board() {
    this(0, 0);
  }
  
  public Board(long contents, int kingmask) {
    this.contents = contents;
    this.kingmask = kingmask;
    if (!initialized) {
      initialized = true;
      initializeMovePermissions();
    }
    instancecount++;
  }
  
  public static void printInstanceCount() {
    System.out.println(instancecount + " copies of board created thus far.");
  }
  
  public void resetBoard() {
    this.contents = -6148915424238676651L; // initial layout... scary, I know
    this.kingmask = 0;
  }
  
  public long getContents() {return contents;}
  public int getKingmask() {return kingmask;}
  
  public int getCellColor(int cell) {
    int retval = (int)(contents >> (cell*2));
    retval = retval & 3; // isolate least significant two bits
    return retval;
  }
  
  public boolean getCellKing(int cell) {
    int val = (kingmask >> cell) & 1;
    return val == 1;
  }
  
  public void setCellColor(int cell, long cellContents) {
    // constrain
    if (cellContents < 0) cellContents = 0;
    if (cellContents > 2) cellContents = 2;
    
    // shortcircuit
    int oldval = getCellColor(cell);
    if (oldval == (int)cellContents) return;
    
    // separate bits
    boolean sethi = (cellContents & 2) == 2;
    boolean setlo = (cellContents & 1) == 1;
    long hibit = 1L << (cell*2+1);
    long lobit = 1L << (cell*2);
    long oldhibit = (oldval & 2) >> 1;
    long oldlobit = oldval & 1;
    
    // debugging
    /*
    System.out.println(sethi);
    System.out.println(setlo);
    System.out.println(Long.toBinaryString(hibit));
    System.out.println(Long.toBinaryString(lobit));
    System.out.println(Long.toBinaryString(oldhibit));
    System.out.println(Long.toBinaryString(oldlobit));
    */
    
    // assign if work needs to be done
    if (oldhibit == 1 && !sethi) this.contents &= ~(hibit);
    else if (oldhibit == 0 && sethi) this.contents |= hibit; 
    if (oldlobit == 1 && !setlo) this.contents &= ~(lobit);
    else if (oldlobit == 0 && setlo) this.contents |= lobit;
  }
  
  public void setCellKing(int cell, boolean king) {
    // shortcircuit
    boolean oldval = getCellKing(cell);
    if (oldval == king) return;
    
    // do work
    int val = 1 << cell;
    if (oldval) this.kingmask &= ~val;
    else this.kingmask |= val;
  }
  
  public void printBoard() {
    //System.out.println("Board state: " + Long.toBinaryString(contents));
    //System.out.println("Kingmask: " + Integer.toBinaryString(kingmask));
    System.out.println("Raw state: " + contents);
    System.out.println("Raw king: " + kingmask);
    System.out.println("red piece count: " + colorCount(RED));
    System.out.println("red king count: " + kingCount(RED));
    System.out.println("black piece count: " + colorCount(BLACK));
    System.out.println("black king count: " + kingCount(BLACK));
    for (int i = 0; i < 32; i++) {
      if (i % 4 == 0) System.out.println("\n+--+--+--+--+--+--+--+--+");
      if (i % 8 == 0) System.out.print("|--|");
      if (i % 8 == 4) System.out.print("|");
      int val = getCellColor(i);
      boolean king = getCellKing(i);
      if (val == EMPTY) System.out.print("  |");
      if (val == ERROR) System.out.print("!!|");
      if (val == RED && !king) System.out.print("rr|");
      if (val == RED && king) System.out.print("RR|");
      if (val == BLACK && !king) System.out.print("bb|");
      if (val == BLACK && king) System.out.print("BB|");
      if (i % 8 != 3) System.out.print("--|");
    }
    System.out.println("\n+--+--+--+--+--+--+--+--+");
    System.out.println();
  }
  
  
    /* neighbor grid for A type cells
    -9       -7
      -4   -3
         X 
      +4   +5
    +7       +9
  */
  
  /* neighbor grid for B type cells
    -9       -7
      -5   -4
         X
      +3   +4
    +7       +9
  */
  
  // black moves hi->low unkinged, red moves low->hi unkinged
  
  
  // here there be dragons: you have been warned!
  public boolean canMove(int cell1, int cell2) {
    boolean isAtype = (cell1 % 8 <= 3); // is the cell an "A" type cell?
    int dist = cell2 - cell1;
    // short circuit (impossible states)
    if (!Arrays.asList(CAN_MOVE_LIST.get(cell1)).contains(cell2)) return false;    
    
    // realm of possibility - is the destination occupied?
    int color2 = getCellColor(cell2);
    if (color2 != EMPTY) {
      //System.out.println("occupied: " + cell2);
      return false;
    } // occupied.
    
    // realm of possibility - destination is empty
    int color1 = getCellColor(cell1);
    boolean king1 = getCellKing(cell1);
    if (color1 == BLACK && (dist < 0 || king1 == KING)) {
      // destination empty, legal mobility (black)
      //System.out.println("legal mobility -> " + cell2);
      if (Math.abs(dist) == 3 || Math.abs(dist) == 4 || Math.abs(dist) == 5) return true; // move to next cell
      // jumping... legal?
      int cellm = EMPTY;
      if (isAtype) {
        switch (dist) {
          case -9:
            cellm = getCellColor(cell1 - 4);
            break;
          case -7:
            cellm = getCellColor(cell1 - 3);
            break;
          case 7:
            cellm = getCellColor(cell1 + 4);
            break;
          case 9:
            cellm = getCellColor(cell1 + 5);
            break;
          default:
            System.out.println("shouldn't get here!");
            return false;
        }
      } else {
        switch (dist) {
          case -9:
            cellm = getCellColor(cell1 - 5);
            break;
          case -7:
            cellm = getCellColor(cell1 - 4);
            break;
          case 7:
            cellm = getCellColor(cell1 + 3);
            break;
          case 9:
            cellm = getCellColor(cell1 + 4);
            break;
          default:
            System.out.println("shouldn't get here!");
            return false;
        }
      }
      //System.out.println("cellm = " + cellm + " for " + cell2);
      if (cellm != RED) return false; // can't jump non-existent piece or same color
      else return true;
    }
    if (color1 == RED && (dist > 0 || king1 == KING)) {
      // destination empty, legal mobility (red)
      if (Math.abs(dist) == 3 || Math.abs(dist) == 4 || Math.abs(dist) == 5) return true; // move to next cell
      // jumping... legal?
      int cellm;
      if (isAtype) {
        switch (dist) {
          case -9:
            cellm = getCellColor(cell1 - 4);
            break;
          case -7:
            cellm = getCellColor(cell1 - 3);
            break;
          case 7:
            cellm = getCellColor(cell1 + 4);
            break;
          case 9:
            cellm = getCellColor(cell1 + 5);
            break;
          default:
            System.out.println("shouldn't get here!");
            return false;
        }
      } else {
        switch (dist) {
          case -9:
            cellm = getCellColor(cell1 - 5);
            break;
          case -7:
            cellm = getCellColor(cell1 - 4);
            break;
          case 7:
            cellm = getCellColor(cell1 + 3);
            break;
          case 9:
            cellm = getCellColor(cell1 + 4);
            break;
          default:
            System.out.println("shouldn't get here!");
            return false;
        }
      }
      //System.out.println("cellm = " + cellm + " for " + cell2);
      if (cellm != BLACK) return false; // can't jump non-existent piece or same color
      else return true;
    }
    
    // no true condition met, so must have failed
    return false;
  }
  
  public String possiblePieces(int color) {
    StringBuilder str = new StringBuilder();
    for (Integer i : getCellsOfColor(color)) 
      if (possibleMovesForCell(i).size() != 0) 
        str.append(i + " ");
    return str.toString().trim();
  }
  
  public ArrayList<Integer> possibleMovesForCell(int cell) {
    ArrayList<Integer> moves = new ArrayList<Integer>();
    for (Integer i : Arrays.asList(CAN_MOVE_LIST.get(cell))) // prevent reduplicated effort
      if (canMove(cell,i)) moves.add(i);
    return moves;
  }
  
  public int colorCount(int color) {
    int count = 0;
    for (int i = 0; i < 32; i++)
      if (getCellColor(i) == color) count++;
    return count;
  }
  
  public int kingCount(int color) {
    int count = 0;
    for (int i = 0; i < 32; i++) 
      if (getCellColor(i) == color && getCellKing(i) == KING) count++;
    return count;
  }
  
  public ArrayList<Integer> getCellsOfColor(int color) {
    ArrayList<Integer> retval = new ArrayList<Integer>();
    for (int i = 0; i < 32; i++) 
      if (getCellColor(i) == color)
        retval.add(i);
    return retval;
  }
  
  public Board move(int cell1, int cell2) {
    if (!canMove(cell1, cell2)) return null; // can't move, return null
    
    Board newboard = new Board(this.contents, this.kingmask);
    boolean isAtype = (cell1 % 8 <= 3); // is the cell an "A" type cell?
    int dist = cell2 - cell1;
    int cell1color = newboard.getCellColor(cell1);
    boolean cell1king = newboard.getCellKing(cell1);
    newboard.setCellColor(cell1, EMPTY);
    newboard.setCellKing(cell1, PAWN);
    newboard.setCellColor(cell2, cell1color);
    newboard.setCellKing(cell2, cell1king);
    if ((cell1color == BLACK && cell2 >= 0 && cell2 <= 3) ||
        (cell1color == RED && cell2 >= 28 && cell2 <= 31))
      newboard.setCellKing(cell2, KING);
    
    if (Math.abs(dist) == 7 || Math.abs(dist) == 9) { // jumping, it gets ugly
      if(isAtype) {
        int middlecell = 0;
        switch(dist) {
          case -9:
            middlecell = cell1 - 4;
            break;
          case -7:
            middlecell = cell1 - 3;
            break;
          case 7:
            middlecell = cell1 + 4;
            break;
          case 9:
            middlecell = cell1 + 5;
            break;
          default:
            System.out.println("Something went wrong here...");
            break;
        }
        newboard.setCellColor(middlecell, EMPTY);
        newboard.setCellKing(middlecell, PAWN);
      } else {
        int middlecell = 0;
        switch(dist) {
          case -9:
            middlecell = cell1 - 5;
            break;
          case -7:
            middlecell = cell1 - 4;
            break;
          case 7:
            middlecell = cell1 + 3;
            break;
          case 9:
            middlecell = cell1 + 4;
            break;
          default:
            System.out.println("Something went wrong here...");
            break;
        }
        newboard.setCellColor(middlecell, EMPTY);
        newboard.setCellKing(middlecell, PAWN);
      }
    }
    return newboard;
  }
  
  //**************************************************************************//
  // utilities                                                                //
  //**************************************************************************//
  
  public static boolean tempCanMove(long contents, int kingmask, int cell1, int cell2) {
    boolean isAtype = (cell1 % 8 <= 3); // is the cell an "A" type cell?
    int dist = cell2 - cell1;
    // short circuit (impossible states)   
    boolean escape = true;
    Integer[] possibles = CAN_MOVE_LIST.get(cell1);
    for(int i = 0; i < possibles.length; i++) {
      if (possibles[i] == cell2) escape = false;
    }
    if(escape) return false;
    
    // realm of possibility - is the destination occupied?
    int color2 = tempGetCellColor(contents, cell2);
    if (color2 != EMPTY) {
      //System.out.println("occupied: " + cell2);
      return false;
    } // occupied.
    
    // realm of possibility - destination is empty
    int color1 = tempGetCellColor(contents, cell1);
    boolean king1 = tempGetCellKing(kingmask, cell1);
    if (color1 == BLACK && (dist < 0 || king1 == KING)) {
      // destination empty, legal mobility (black)
      //System.out.println("legal mobility -> " + cell2);
      if (Math.abs(dist) == 3 || Math.abs(dist) == 4 || Math.abs(dist) == 5) return true; // move to next cell
      // jumping... legal?
      int cellm = EMPTY;
      if (isAtype) {
        switch (dist) {
          case -9:
            cellm = tempGetCellColor(contents, cell1 - 4);
            break;
          case -7:
            cellm = tempGetCellColor(contents, cell1 - 3);
            break;
          case 7:
            cellm = tempGetCellColor(contents, cell1 + 4);
            break;
          case 9:
            cellm = tempGetCellColor(contents, cell1 + 5);
            break;
          default:
            System.out.println("shouldn't get here!");
            return false;
        }
      } else {
        switch (dist) {
          case -9:
            cellm = tempGetCellColor(contents, cell1 - 5);
            break;
          case -7:
            cellm = tempGetCellColor(contents, cell1 - 4);
            break;
          case 7:
            cellm = tempGetCellColor(contents, cell1 + 3);
            break;
          case 9:
            cellm = tempGetCellColor(contents, cell1 + 4);
            break;
          default:
            System.out.println("shouldn't get here!");
            return false;
        }
      }
      //System.out.println("cellm = " + cellm + " for " + cell2);
      if (cellm != RED) return false; // can't jump non-existent piece or same color
      else return true;
    }
    if (color1 == RED && (dist > 0 || king1 == KING)) {
      // destination empty, legal mobility (red)
      if (Math.abs(dist) == 3 || Math.abs(dist) == 4 || Math.abs(dist) == 5) return true; // move to next cell
      // jumping... legal?
      int cellm;
      if (isAtype) {
        switch (dist) {
          case -9:
            cellm = tempGetCellColor(contents, cell1 - 4);
            break;
          case -7:
            cellm = tempGetCellColor(contents, cell1 - 3);
            break;
          case 7:
            cellm = tempGetCellColor(contents, cell1 + 4);
            break;
          case 9:
            cellm = tempGetCellColor(contents, cell1 + 5);
            break;
          default:
            System.out.println("shouldn't get here!");
            return false;
        }
      } else {
        switch (dist) {
          case -9:
            cellm = tempGetCellColor(contents, cell1 - 5);
            break;
          case -7:
            cellm = tempGetCellColor(contents, cell1 - 4);
            break;
          case 7:
            cellm = tempGetCellColor(contents, cell1 + 3);
            break;
          case 9:
            cellm = tempGetCellColor(contents, cell1 + 4);
            break;
          default:
            System.out.println("shouldn't get here!");
            return false;
        }
      }
      //System.out.println("cellm = " + cellm + " for " + cell2);
      if (cellm != BLACK) return false; // can't jump non-existent piece or same color
      else return true;
    }
    
    // no true condition met, so must have failed
    return false;
  }
  
  public static String tempPossiblePieces(long contents, int kingmask, int color) {
    StringBuilder str = new StringBuilder();
    for (Integer i : tempGetCellsOfColor(contents, color)) 
      if (tempPossibleMovesForCell(contents, kingmask, i).size() != 0) 
        str.append(i + " ");
    return str.toString().trim();
  }
  
  public static int tempGetCellColor(long contents, int cell) {
    int retval = (int)(contents >> (cell*2));
    retval = retval & 3; // isolate least significant two bits
    return retval;
  }
  
  public static boolean tempGetCellKing(int kingmask, int cell) {
    int val = (kingmask >> cell) & 1;
    return val == 1;
  }
  
  public static long tempSetCellColor(long contents, int cell, long cellContents) {
    long temp = contents;
    // constrain
    if (cellContents < 0) cellContents = 0;
    if (cellContents > 2) cellContents = 2;
    
    // shortcircuit
    int oldval = tempGetCellColor(contents, cell);
    if (oldval == (int)cellContents) return contents;
    
    // separate bits
    boolean sethi = (cellContents & 2) == 2;
    boolean setlo = (cellContents & 1) == 1;
    long hibit = 1L << (cell*2+1);
    long lobit = 1L << (cell*2);
    long oldhibit = (oldval & 2) >> 1;
    long oldlobit = oldval & 1;
    
    // assign if work needs to be done
    if (oldhibit == 1 && !sethi) temp &= ~(hibit);
    else if (oldhibit == 0 && sethi) temp |= hibit; 
    if (oldlobit == 1 && !setlo) temp &= ~(lobit);
    else if (oldlobit == 0 && setlo) temp |= lobit;
    
    return temp;
  }
  
  public static int tempSetCellKing(int kingmask, int cell, boolean king) {
    int temp = kingmask;
    // shortcircuit
    boolean oldval = tempGetCellKing(kingmask, cell);
    if (oldval == king) return temp;
    
    // do work
    int val = 1 << cell;
    if (oldval) temp &= ~val;
    else temp |= val;
    
    return temp;
  }
  
  public static Long tempMoveCells(long contents, int kingmask, int cell1, int cell2) {
    long retval = 0;
    if (!tempCanMove(contents, kingmask, cell1, cell2)) return null; // can't move, return null
    boolean isAtype = (cell1 % 8 <= 3); // is the cell an "A" type cell?
    int dist = cell2 - cell1;
    int cell1color = tempGetCellColor(contents, cell1);
    retval = tempSetCellColor(tempSetCellColor(contents, cell2, cell1color), cell1, EMPTY);
    
    if (Math.abs(dist) == 7 || Math.abs(dist) == 9) { // jumping, it gets ugly
      if(isAtype) {
        int middlecell = 0;
        switch(dist) {
          case -9:
            middlecell = cell1 - 4;
            break;
          case -7:
            middlecell = cell1 - 3;
            break;
          case 7:
            middlecell = cell1 + 4;
            break;
          case 9:
            middlecell = cell1 + 5;
            break;
          default:
            System.out.println("Something went wrong here...");
            break;
        }
        retval = tempSetCellColor(retval, middlecell, EMPTY);
      } else {
        int middlecell = 0;
        switch(dist) {
          case -9:
            middlecell = cell1 - 5;
            break;
          case -7:
            middlecell = cell1 - 4;
            break;
          case 7:
            middlecell = cell1 + 3;
            break;
          case 9:
            middlecell = cell1 + 4;
            break;
          default:
            System.out.println("Something went wrong here...");
            break;
        }
        retval = tempSetCellColor(retval, middlecell, EMPTY);
      }
    }
    return retval;
  }    
  
  public static Integer tempMoveKing(long contents, int kingmask, int cell1, int cell2) {
    int retval = 0;
    if (!tempCanMove(contents, kingmask, cell1, cell2)) return null; // can't move, return null
    boolean isAtype = (cell1 % 8 <= 3); // is the cell an "A" type cell?
    int dist = cell2 - cell1;
    int cell1color = tempGetCellColor(contents, cell1);
    boolean cell1king = tempGetCellKing(kingmask, cell1);
    retval = tempSetCellKing(tempSetCellKing(kingmask, cell2, PAWN), cell1, cell1king);
    if ((cell1color == BLACK && cell2 >= 0 && cell2 <= 3) ||
        (cell1color == RED && cell2 >= 28 && cell2 <= 31))
      retval = tempSetCellKing(retval, cell2, KING);
    
    if (Math.abs(dist) == 7 || Math.abs(dist) == 9) { // jumping, it gets ugly
      if(isAtype) {
        int middlecell = 0;
        switch(dist) {
          case -9:
            middlecell = cell1 - 4;
            break;
          case -7:
            middlecell = cell1 - 3;
            break;
          case 7:
            middlecell = cell1 + 4;
            break;
          case 9:
            middlecell = cell1 + 5;
            break;
          default:
            System.out.println("Something went wrong here...");
            break;
        }
        retval = tempSetCellKing(retval, middlecell, PAWN);
      } else {
        int middlecell = 0;
        switch(dist) {
          case -9:
            middlecell = cell1 - 5;
            break;
          case -7:
            middlecell = cell1 - 4;
            break;
          case 7:
            middlecell = cell1 + 3;
            break;
          case 9:
            middlecell = cell1 + 4;
            break;
          default:
            System.out.println("Something went wrong here...");
            break;
        }
        retval = tempSetCellKing(retval, middlecell, PAWN);
      }
    }
    return retval;
  }
  
  public static int tempColorCount(long contents, int color) {
    int count = 0;
    for (int i = 0; i < 32; i++)
      if (tempGetCellColor(contents, i) == color) count++;
    return count;
  }
  
  public static int tempKingCount(long contents, int kingmask, int color) {
    int count = 0;
    for (int i = 0; i < 32; i++) 
      if (tempGetCellColor(contents, i) == color && tempGetCellKing(kingmask, i) == KING) count++;
    return count;
  }
  
  public static ArrayList<Integer> tempGetCellsOfColor(long contents, int color) {
    ArrayList<Integer> retval = new ArrayList<Integer>();
    for (int i = 0; i < 32; i++) 
      if (tempGetCellColor(contents, i) == color)
        retval.add(i);
    return retval;
  }

  public static ArrayList<Integer> tempPossibleMovesForCell(long contents, int kingmask, int cell) {
    ArrayList<Integer> moves = new ArrayList<Integer>();
    Integer[] possibles = CAN_MOVE_LIST.get(cell);
    for (int i = 0; i < possibles.length; i++) // prevent reduplicated effort
      if (tempCanMove(contents, kingmask, cell,possibles[i])) moves.add(possibles[i]);
    return moves;
  }

}
