// Game class
// CSC 375 - Lea
// Jacob Peck

// to be run on lambda:
// java Game auto

import java.util.*;
import java.net.*;
import java.io.*;

public class Game {
  private static boolean auto = false;
  private static boolean humanred = false;
  private static boolean humanblack = false;
  
  public static Socket socket;
  public static ObjectInputStream in;
  public static ObjectOutputStream out;
  
  public static void main(String[] args) {
    try{
      socket = new Socket("rho",3553);
      out = new ObjectOutputStream(socket.getOutputStream());
      in = new ObjectInputStream(socket.getInputStream());
    } catch (Exception ex) {ex.printStackTrace(); System.exit(-1);}
    
    if (args.length >= 1) {
      if (Arrays.asList(args).contains("auto")) auto = true;
      if (Arrays.asList(args).contains("humanred")) humanred = true;
      if (Arrays.asList(args).contains("humanblack")) humanblack = true;
    }
    Board gameboard = new Board();
    Player red = new Player();
    Player black = new Player(true);
    Scanner sc = new Scanner(System.in);
    
    gameboard.resetBoard();
    
    System.out.println("Checkers!... kind of.");
    System.out.println("Rules changes: no multiple jumps, and jumps are optional.");
    System.out.println("Red player plays first.");
    
    gameboard.printBoard();
    
    int turnswithoutcarnage = 0;
    
    if (!auto) parseInput(sc);
    
    // gameloop
    while(gameboard != null && gameboard.colorCount(Board.RED) != 0 && 
          gameboard.colorCount(Board.BLACK) != 0 && turnswithoutcarnage < 50) {
      int oldredcount = gameboard.colorCount(Board.RED);
      int oldblackcount = gameboard.colorCount(Board.BLACK);
      
      if (!humanred) {
        System.out.println("RED: Thinking...");
        gameboard = red.makeMove(gameboard);
        Board.printInstanceCount();
        FindMove.printInstanceCount();
      } else {
        System.out.println("Pieces that can move: " + gameboard.possiblePieces(Board.RED));
        gameboard = parseForMove(sc,gameboard,false);
      }
      
      if (gameboard != null) gameboard.printBoard();
      else {System.out.println("Red loses, can't make move!"); break;}
      if (!auto) parseInput(sc);
      
      if (gameboard.colorCount(Board.BLACK) == 0) break;
      
      if (gameboard.colorCount(Board.BLACK) == oldblackcount)
        turnswithoutcarnage++;
      else
        turnswithoutcarnage = 0;
      oldblackcount = gameboard.colorCount(Board.BLACK);
      
      if (!humanblack) {
        System.out.println("BLACK: Thinking...");
        gameboard = black.makeMove(gameboard);
        Board.printInstanceCount();
        FindMove.printInstanceCount();
      } else {
        System.out.println("Pieces that can move: " + gameboard.possiblePieces(Board.BLACK));
        gameboard = parseForMove(sc,gameboard,true);
      }
      
      if (gameboard != null) gameboard.printBoard();
      else {System.out.println("Black loses, can't make move!"); break;}
      if (!auto) parseInput(sc);
      
      if (gameboard.colorCount(Board.RED) == oldredcount)
        turnswithoutcarnage++;
      else
        turnswithoutcarnage = 0;
    }
    
    if (turnswithoutcarnage >= 50)
      System.out.println("Stalemate... " + turnswithoutcarnage + " turns have passed without material loss.");
    else if (gameboard == null)
      System.out.println("Can't make move, game over.");
    else if(gameboard.colorCount(Board.RED) == 0)
      System.out.println("Black wins!");
    else
      System.out.println("Red wins!");
     
    Board.printInstanceCount();
    try {
      out.flush();
      out.close();
      socket.close();
    } catch (Exception ex) {}
    System.exit(0);
  }
  
  public static void parseInput(Scanner sc) {
    System.out.print("?> ");
    String str = sc.nextLine();
    if (str.equalsIgnoreCase("auto")) auto = true;
    if (str.equalsIgnoreCase("quit")) System.exit(0);
  }
  
  public static Board parseForMove(Scanner sc,Board gameboard, boolean isBlack) {
    boolean success = false;
    int cell = 0;
    int move = 0;
    while (!success) {
      sc = new Scanner(System.in);
      System.out.print("Cell to move from? ");
      try {
        cell = sc.nextInt();
      } catch (Exception ex) {
        System.out.println("Try again.");
        continue;
      }
      if(isBlack && gameboard.getCellColor(cell) != Board.BLACK) {
        System.out.println("You're playing BLACK!");
        continue;
      }
      if(!isBlack && gameboard.getCellColor(cell) != Board.RED) {
        System.out.println("You're playing RED!");
        continue;
      }
      System.out.println("That piece may move to: " + gameboard.possibleMovesForCell(cell));
      System.out.print("Cell to move to? ");
      try {
        move = sc.nextInt();
      } catch (Exception ex) {
        System.out.println("Try again.");
        continue;
      }
      
      if (!gameboard.canMove(cell,move))
        System.out.println("Can't make that move, sorry.  Try again.");
      else
        success = true;
    }
    
    return gameboard.move(cell,move);
  }
}
