// Player class
// CSC 375 - Lea
// Jacob Peck


import java.util.*;
import java.util.concurrent.*;

public class Player {
  private Board board;
  private boolean isBlack = false;
  public static final Random random = new Random(); // for variability
  
  public static final ForkJoinPool fp = new ForkJoinPool();
  
  public Player() {this(false);}
  public Player(boolean isBlack) {
     this.isBlack = isBlack;
  }
  
  // concurrent
  public Board makeMove(Board newboard) {
    this.board = newboard;
    int color; int othercolor;
    if(isBlack) {color = Board.BLACK; othercolor = Board.RED;}
    else {color = Board.RED; othercolor = Board.BLACK;}
    
    // temporary storage for move
    int sef = 0;
    int cell1 = 0;
    int cell2 = 0;
    
    // determine move
    for(Integer piece1 : this.board.getCellsOfColor(color)) {
      //System.out.println("Evaluating piece: " + piece1);
      HashMap<Integer, FindMove> tasklist = new HashMap<Integer, FindMove>();
      for(Integer move1 : this.board.possibleMovesForCell(piece1)) {
        Long contents = Board.tempMoveCells(this.board.getContents(),
                                            this.board.getKingmask(), piece1, move1);
        Integer kingmask = Board.tempMoveKing(this.board.getContents(), 
                                          this.board.getKingmask(), piece1, move1);
        if(contents == null || kingmask == null) continue;
        FindMove fm = new FindMove(contents, kingmask, isBlack, 1);
        tasklist.put(move1,fm);
        //fp.invoke(fm);
      }
      int itercount = 0;
      for(Map.Entry<Integer, FindMove> entry : tasklist.entrySet()) {
        FindMove fm = entry.getValue();
        if (itercount % 2 != 0) {
          try {
            Game.out.writeObject(fm);
            Game.out.flush();
          } catch (Exception ex) { ex.printStackTrace(); System.exit(-1); }
        } else {
          fp.invoke(fm);
        }
        
        int move1 = entry.getKey();
        int newval = 0;
        
        if (itercount % 2 != 0) {
          try{
            newval = Game.in.readInt();
          } catch (Exception ex) { ex.printStackTrace(); System.exit(-1); }
        } else {
          newval = fm.join();
        }
        
        fm = null; // allow to GC
        if (isBlack) {
          if (sef == 0 || newval < sef || (newval == sef && random.nextDouble() > .25)) {
            sef = newval;
            cell1 = piece1;
            cell2 = move1;
            //System.out.println("Setting values: " + cell1 + " -> " + cell2);
          }
        } else {
          if (sef == 0 || newval > sef || (newval == sef && random.nextDouble() > .25)) {
            sef = newval;
            cell1 = piece1;
            cell2 = move1;
            //System.out.println("Setting values: " + cell1 + " -> " + cell2);
          }
        }
        itercount++;
      }
    }
    //System.gc();
    
    //System.out.println("making move " + cell1 + " -> " + cell2);
    this.board = board.move(cell1,cell2);
    
    return this.board;
  }
}

class FindMove extends RecursiveTask<Integer> implements java.io.Serializable {
  final int kingmask;
  final long contents;
  final boolean isBlack;
  final int level;
  final int color;
  final static int DEPTH = 7;
  private static int instancecount = 0;
  FindMove(long contents, int kingmask, boolean isBlack, int level) {
    this.contents = contents;
    this.kingmask = kingmask;
    this.isBlack = isBlack;
    this.level = level;
    if (isBlack) this.color = Board.BLACK;
    else this.color = Board.RED;
    instancecount++;
  }
  public static void printInstanceCount() {
    System.out.println(instancecount + " instances of FindMove created thus far.");
  }
  public Integer compute() {
    //System.out.println("  Computing... level " + level + " isBlack? " + isBlack);
    int sef = 0;
    if (level < DEPTH) {      
      ArrayList<FindMove> tasklist = new ArrayList<FindMove>();
      for (Integer cell : Board.tempGetCellsOfColor(contents, color))
        for (Integer i : Board.tempPossibleMovesForCell(contents, kingmask, cell)) {
          FindMove x = new FindMove(Board.tempMoveCells(contents, kingmask, cell,i),
                                    Board.tempMoveKing(contents, kingmask, cell, i), 
                                    !this.isBlack, level+1);
          tasklist.add(x);
          x.fork();
          //System.out.println("    contemplating move: " + cell + " -> " + i);
        }
      Collections.reverse(tasklist); // keep the F/J framework working
      for (FindMove fm : tasklist) {
        int newval = fm.join();
        fm = null; // allow to GC
        if (sef == 0 || newval > sef || (newval == sef && Player.random.nextDouble() > .25)) sef = newval;
      }
    }
    if (Board.tempColorCount(contents, Board.BLACK) == 0) return sef + 1000000;
    if (Board.tempColorCount(contents, Board.RED) == 0) return sef - 1000000;
    int red = Board.tempColorCount(contents, Board.RED) 
            + Board.tempKingCount(contents, kingmask, Board.RED);
    int black = Board.tempColorCount(contents, Board.BLACK) 
              + Board.tempKingCount(contents, kingmask, Board.BLACK);
    return sef + (red - black);
  }
}
